import numpy as np
import cv2
import torch
import torch.nn.functional as F
import time
import torchvision.transforms as transforms

from network.SiamFCOS import SiamFCOS
from config import config
from utils.sample_space_model import GMM
from utils.custom_transforms import ToTensor
from utils.utils import get_exemplar_image, get_instance_image, get_instance_image_detect, box_transform_inv,add_box_img,add_box_img_left_top
from utils.generate_anchors import generate_anchors
from collections import OrderedDict
if config.use_gpu:
    import cupy as cp
from copy import copy
torch.set_num_threads(1)  # otherwise pytorch will take all cpus


class SiamFCOSTracker:
    def __init__(self, model_path):
        self.num_select = 3
        self._num_feature_blocks = 3
        self.model = SiamFCOS()
        checkpoint = torch.load(model_path)
        if 'network' in list(checkpoint.keys()):
            self.model.load_state_dict(torch.load(model_path)['network'])
        else:
            self.model.load_state_dict(torch.load(model_path))
        self.model = self.model.cuda()
        self.model.eval()
        # self.transforms = transforms.Compose(
        #     [transforms.ToTensor()]
        # )
        self.transforms = transforms.Compose([ToTensor()])

        # valid_scope = 2 * config.valid_scope + 1
        self.anchors = generate_anchors(config.total_stride, config.anchor_base_size, config.anchor_scales,
                                        config.anchor_ratios,
                                        config.response_map_size)
        self.window = np.tile(np.outer(np.hanning(25), np.hanning(25))[None, :],
                              [config.anchor_num, 1, 1]).flatten()

        self.bbox_x_list = []
        self.bbox_y_list = []
        self.displacement_center_x = []
        self.displacement_center_y = []
        self.displacement_mean_y = 0
        self.displacement_mean_x = 0

    def _cosine_window(self, size):
        """
            get the cosine window
        """
        cos_window = np.hanning(int(size[0]))[:, np.newaxis].dot(np.hanning(int(size[1]))[np.newaxis, :])
        cos_window = cos_window.astype(np.float32)
        cos_window /= np.sum(cos_window)
        return cos_window

    def init(self, frame, bbox):
        """ initialize siamfc tracker
        Args:
            frame: an RGB image
            bbox: one-based bounding box [x, y, width, height]
        """

        self.count = 1
        self.score = []
        xp = np
        # xp = cp if config.use_gpu else np
        self.pos = np.array(
            [bbox[0] + bbox[2] / 2 - 1 / 2, bbox[1] + bbox[3] / 2 - 1 / 2])  # center x, center y, zero based
        # self.pos = np.array(
        #     [bbox[0] + bbox[2] // 2, bbox[1] + bbox[3] // 2])  # same to original code
        self.target_sz = np.array([bbox[2], bbox[3]])  # width, height
        self.bbox = np.array([bbox[0] + bbox[2] / 2 - 1 / 2, bbox[1] + bbox[3] / 2 - 1 / 2, bbox[2], bbox[3]])
        # self.bbox = np.array(
        #     [bbox[0] + bbox[2] // 2, bbox[1] + bbox[3] // 2, bbox[2], bbox[3]])  # same to original code
        self.origin_target_sz = np.array([bbox[2], bbox[3]])
        # get exemplar img
        self.img_mean = np.mean(frame, axis=(0, 1))
        self._num_samples = config.num_samples
        self.yolo = Detect()

        exemplar_img, scale_z, _ = get_exemplar_image(frame, self.bbox,
                                                config.exemplar_size, config.context_amount, self.img_mean)

        instance_img_np, _, _, scale_x = get_instance_image(frame, self.bbox, config.exemplar_size,
                                                            config.instance_size,
                                                            config.context_amount, self.img_mean)

        exemplar_img = self.transforms(exemplar_img)[None, :, :, :]
        # GMM
        self._gmm = GMM(self._num_samples)
        self._samplesf = [[]] * self._num_feature_blocks
        self._num_training_samples = 0

        for i in range(self._num_feature_blocks):
            if not config.use_gpu:
                self._samplesf[i] = np.zeros((int(config.feature_size), int(config.feature_size),
                                              config.feature_channel, config.num_samples), dtype=np.float32)
            else:
                self._samplesf[i] = cp.zeros((int(config.feature_size), int(config.feature_size),
                                              config.feature_channel, config.num_samples), dtype=cp.float32)


        self._num_training_samples = 0

        examplar_output = self.model.examplar_feature_extract(exemplar_img.cuda())
        self.first_feature = examplar_output

        for i in range(self._num_feature_blocks):
            tmp = examplar_output[i].permute(2, 3, 1, 0).cpu().detach().numpy()
            self._samplesf[i][:, :, :, 0: 1] = tmp

        merged_sample, new_sample, merged_sample_id, new_sample_id, self.weight = \
            self._gmm.update_sample_space_model(self._samplesf, self._samplesf, self._num_training_samples)

        self._num_training_samples += 1
        self.miss_num = 0

        self.displacement_mean_x = 0
        self.displacement_mean_y = 0
        self.bbox_x_list.append(self.pos[0])
        self.bbox_y_list.append(self.pos[1])

        self.sthresh = []
        # print(self.bbox)

    def tracking(self, instance_img, search_region_output, samples):
        pred_score, pred_centerness, pred_regression = self.model(samples, search_region_output, instance_img)

        pred_conf = pred_score.reshape(-1, 2, 25 * 25).permute(0, 2, 1)
        pred_offset = pred_regression.reshape(-1, 4, 25 * 25).permute(0, 2, 1)
        pred_center = pred_centerness.reshape(-1, 1, 25 * 25).permute(0, 2, 1)

        center = pred_center[0].cpu().detach().numpy()
        center = center.flatten()
        center = (center - min(center)) / (max(center) - min(center))

        delta = pred_offset[0].cpu().detach().numpy()
        box = box_transform_inv(self.anchors, delta)  # cx, cy, w, h
        score_pred = F.softmax(pred_conf, dim=2)[0, :, 1].cpu().detach().numpy()

        score_merge = score_pred * center
        # pscore = pscore * (1 - config.window_influence) + self.window * config.window_influence
        sort_index = np.argsort(score_merge)  # 从小到大排序
        best_pscore_id = sort_index[-self.num_select:]



        return score_pred, center, delta, score_merge, best_pscore_id, box

    def GMM_tracking(self, instance_img, search_region_output):
        p_score = []
        p_centerness = []
        p_regression = []
        for i in range(self._num_training_samples):
            samples = []
            for j in range(self._num_feature_blocks):
                sample = self._samplesf[j][:, :, :, i]
                sample = np.squeeze(sample)
                sample = sample.transpose(2, 0, 1)
                sample = torch.from_numpy(sample)[None, :, :, :]
                samples.append(sample.cuda())

            score_pred, center, delta, score_merge, best_pscore_id, _ = self.tracking(instance_img, search_region_output, samples)

            p_score.append(score_pred)
            p_centerness.append(center)
            p_regression.append(delta)

        if self._num_training_samples > 1:
            score = 0
            cen = 0
            regression = 0
            sample_weight = self.weight / np.sum(self.weight)
            for i in range(len(p_regression)):
                score += sample_weight[i] * p_score[i]
                cen += sample_weight[i] * p_centerness[i]
                regression += sample_weight[i] * p_regression[i]

            score_pred = score
            center = cen
            delta = regression

        box_pred = box_transform_inv(self.anchors, delta)  # cx, cy, w, h

        return score_pred, center, delta, score_merge, best_pscore_id, box_pred

    def update(self, frame):
        """track object based on the previous frame
        Args:
            frame: an RGB image

        Returns:
            bbox: tuple of 1-based bounding box(xmin, ymin, xmax, ymax)
        """
        self.count += 1
        instance_img_np, _, _, scale_x = get_instance_image(frame, self.bbox, config.exemplar_size,
                                                         config.instance_size,
                                                         config.context_amount, self.img_mean)

        instance_img = self.transforms(instance_img_np)[None, :, :, :]
        self.scale_x = scale_x
        search_region_output = self.model.search_region_feature_extract(instance_img.cuda())

        score_pred, center, delta, score_merge, best_pscore_id, box_pred = self.GMM_tracking(instance_img, search_region_output)

        fore_id_5 = np.argwhere(score_pred>0.5)
        fore_id_9 = np.argwhere(score_pred>0.9)

        def change(r):
            return np.maximum(r, 1. / r)

        def sz(w, h):
            pad = (w + h) * 0.5
            sz2 = (w + pad) * (h + pad)
            return np.sqrt(sz2)

        def sz_wh(wh):
            pad = (wh[0] + wh[1]) * 0.5
            sz2 = (wh[0] + pad) * (wh[1] + pad)
            return np.sqrt(sz2)

        # self.target_sz 为目标真实宽高，scale_x为bbox扩大大约四倍，resize到271
        # self.target_sz * scale_x为目标在271的图上的实际宽高
        # sz_wh为扩大大约两倍，变成正方形
        # 对当前目标预测出来的尺寸与上一帧目标尺寸变化进行惩罚，改变越大，惩罚越大
        s_c = change(sz(box_pred[:, 2], box_pred[:, 3]) / (sz_wh(self.target_sz * scale_x)))  # scale penalty
        # 对当前目标预测出来的宽高比例与上一帧目标的宽高比例变化的惩罚
        r_c = change((self.target_sz[0] / self.target_sz[1]) / (box_pred[:, 2] / box_pred[:, 3]))  # ratio penalty
        penalty = np.exp(-(r_c * s_c - 1.) * config.penalty_k)  # config.penalty_k为0.22， 最大为1
        score_merge = score_pred * center
        pscore = penalty * score_merge

        fore_pixels5 = self.anchors[fore_id_5, :2] / scale_x
        fore_pixels9 = self.anchors[fore_id_9, :2] / scale_x

        fore_pixels5 = fore_pixels5 + self.pos
        fore_pixels9 = fore_pixels9 + self.pos


        # 干扰问题cos
        # if number_label > 1:
        #     pscore = pscore * (1 - 0.75) + self.window * 0.75
        #     sort_index = np.argsort(pscore)  # 从小到大排序
        #     best_pscore_id = sort_index[-self.num_select:]
        #     maxscore = pscore[best_pscore_id].mean()
        # else:
        #     pscore = pscore * (1 - config.window_influence) + self.window * config.window_influence
        #     sort_index = np.argsort(pscore)  # 从小到大排序
        #     best_pscore_id = sort_index[-self.num_select:]
        #     maxscore = pscore[best_pscore_id].mean()

        keep = (fore_pixels5[:, :, 0] > (self.bbox[0] - self.bbox[2] / 2 + 0.5)) & (
                fore_pixels5[:, :, 0] < (self.bbox[0] + self.bbox[2] / 2 - 0.5)) & \
               (fore_pixels5[:, :, 1] > (self.bbox[1] - self.bbox[3] / 2 + 0.5)) & (
                fore_pixels5[:, :, 1] < (self.bbox[1] + self.bbox[3] / 2 - 0.5))
        p5_5 = len(fore_pixels5[keep]) / (len(fore_pixels5) + 1e-6)  # 从小到大排序

        # 干扰问题cos
        if p5_5 > 0.25:
            pscore = pscore * (1 - 0.75) + self.window * 0.75
            sort_index = np.argsort(pscore)  # 从小到大排序
            best_pscore_id = sort_index[-self.num_select:]
            maxscore = pscore[best_pscore_id].mean()
        else:
            pscore = pscore * (1 - config.window_influence) + self.window * config.window_influence
            sort_index = np.argsort(pscore)  # 从小到大排序
            best_pscore_id = sort_index[-self.num_select:]
            maxscore = pscore[best_pscore_id].mean()

        # pscore = pscore * (1 - config.window_influence) + self.window * config.window_influence
        # sort_index = np.argsort(pscore)  # 从小到大排序
        # best_pscore_id = sort_index[-self.num_select:]
        # maxscore = pscore[best_pscore_id].mean()

        maxscore_merge = score_merge[best_pscore_id].mean()

        target = box_pred[best_pscore_id, :] / scale_x  # 回到原图尺寸，中心和宽高

        lr = pscore[best_pscore_id] * config.lr_box  # config.lr_box=0.3

        # self.pos[0] = 上一帧绝对中心
        # target是以图像中心为原点的坐标
        res_x = np.clip(target[:, 0] + self.pos[0], 0, frame.shape[1])  #
        res_y = np.clip(target[:, 1] + self.pos[1], 0, frame.shape[0])  #

        # config.min_scale = 0.1， config.min_scale = 10
        res_w = np.clip(self.target_sz[0] * (1 - lr) + target[:, 2] * lr, config.min_scale * self.origin_target_sz[0],
                        config.max_scale * self.origin_target_sz[0])
        res_h = np.clip(self.target_sz[1] * (1 - lr) + target[:, 3] * lr, config.min_scale * self.origin_target_sz[1],
                        config.max_scale * self.origin_target_sz[1])

        res_x = np.mean(res_x, axis=-1)
        res_y = np.mean(res_y, axis=-1)
        res_w = np.mean(res_w, axis=-1)
        res_h = np.mean(res_h, axis=-1)
        bbox = np.array([res_x, res_y, res_w, res_h])

        ###目标检测###
        oj_label_track = []

        flag = 0
        # 目标消失
        # print('frame_id: ', self.count-1)
        # print('maxscore: ', maxscore)
        # print('maxscore_merge: ', maxscore_merge)
        # self.thresh1 = 0.6
        # flag = 0
        if self.count == 2:
            if maxscore_merge < 0.55:
                self.thresh = 0.1
                self.p95_thresh = 0.2
            else:
                self.thresh = 0.55
                self.p95_thresh = 0.5

        if self.thresh == 0.1:
            if self.count < 4:
                self.sthresh.append(maxscore_merge)
            if self.count == 4:
                self.thresh = np.mean(self.sthresh) * 0.8

        if self.thresh != 0.55:
            if maxscore_merge < self.thresh:
                bbox = self.redetect(frame)
                self.miss_num += 1
            else:
                self.miss_num = 0
        else:
            if maxscore < self.thresh :#or maxscore_merge < self.thresh:
                bbox = self.redetect(frame)
                self.miss_num += 1
            else:
                self.miss_num = 0

        # print('miss_num: ', self.miss_num)

        bbox = (np.clip(bbox[0]-bbox[2]/2+0.5,0,frame.shape[1]).astype(np.float64),
            np.clip(bbox[1]-bbox[3]/2+0.5, 0, frame.shape[0]).astype(np.float64),
            np.clip(bbox[0]+bbox[2]/2-0.5, 0, frame.shape[1]).astype(np.float64),
            np.clip(bbox[1]+bbox[3]/2-0.5, 0, frame.shape[0]).astype(np.float64))

        self.bbox = (
            np.clip((bbox[0]+bbox[2])/2, 0, frame.shape[1]).astype(np.float64),
            np.clip((bbox[1]+bbox[3])/2, 0, frame.shape[0]).astype(np.float64),
            np.clip((bbox[2]-bbox[0]+1), 10, frame.shape[1]).astype(np.float64),
            np.clip(bbox[3]-bbox[1]+1, 10, frame.shape[0]).astype(np.float64))
        # print(self.bbox)

        self.pos = np.array([self.bbox[0], self.bbox[1]])
        self.target_sz = np.array([self.bbox[2], self.bbox[3]])

        # print(fore_pixels5[:2])

        if self.count > 1530:
            a = 1
        keep = (fore_pixels9[:, :, 0] > (bbox[0] - bbox[2] / 2 + 0.5)) & (fore_pixels9[:, :, 0] < (bbox[0] + bbox[2] / 2 - 0.5)) & \
               (fore_pixels9[:, :, 1] > (bbox[1] - bbox[3] / 2 + 0.5)) & (fore_pixels9[:, :, 1] < (bbox[1] + bbox[3] / 2 - 0.5))

        fore_pixels9_keep = fore_pixels9[keep]
        p_95 = len(fore_pixels9_keep) / (len(fore_pixels5) + 1e-6)

        if p_95 > self.p95_thresh:
            new_exemplar_img, scale_z, _ = get_exemplar_image(frame, self.bbox,
                                                              config.exemplar_size, config.context_amount,
                                                              self.img_mean)
            new_exemplar_img = self.transforms(new_exemplar_img)[None, :, :, :]
            new_exemplar_sample = self.model.examplar_feature_extract(new_exemplar_img.cuda())

            new_samples = [[]] * self._num_feature_blocks
            for i in range(self._num_feature_blocks):
                new_samples[i] = np.zeros((int(config.feature_size), int(config.feature_size),
                                           config.feature_channel, 1), dtype=np.float32)
                sample = new_exemplar_sample[i]
                sample = sample.permute(2, 3, 1, 0).cpu().detach().numpy()
                new_samples[i] = sample

            merged_sample, new_sample, merged_sample_id, new_sample_id, self.weight = \
                self._gmm.update_sample_space_model(self._samplesf, new_samples, self._num_training_samples)

            if config.learning_rate > 0:
                for i in range(self._num_feature_blocks):
                    if merged_sample_id >= 0:
                        self._samplesf[i][:, :, :, merged_sample_id:merged_sample_id + 1] = merged_sample[i]
                    if new_sample_id >= 0:
                        self._samplesf[i][:, :, :, new_sample_id:new_sample_id + 1] = new_sample[i]

            if self._num_training_samples < self._num_samples:
                self._num_training_samples += 1


        self.bbox_x_list.append(self.pos[0])
        self.bbox_y_list.append(self.pos[1])

        if len(self.bbox_x_list) == 6 & len(self.bbox_y_list) == 6:
            self.bbox_x_list.remove(self.bbox_x_list[0])
            self.bbox_y_list.remove(self.bbox_y_list[0])

        if len(self.bbox_x_list) == 5 & len(self.bbox_y_list) == 5:
            self.displacement_center_x.append(self.bbox_x_list[1] - self.bbox_x_list[0])
            self.displacement_center_x.append(self.bbox_x_list[2] - self.bbox_x_list[1])
            self.displacement_center_x.append(self.bbox_x_list[3] - self.bbox_x_list[2])
            self.displacement_center_x.append(self.bbox_x_list[4] - self.bbox_x_list[3])
            me_x = 1/4*(self.displacement_center_x[0] + self.displacement_center_x[1] + self.displacement_center_x[2] + \
                        self.displacement_center_x[3])
            if self.displacement_center_x[3]*me_x > 0:
                self.displacement_mean_x = me_x
            else:
                self.displacement_mean_x = 0
            self.displacement_center_y.append(self.bbox_y_list[1] - self.bbox_y_list[0])
            self.displacement_center_y.append(self.bbox_y_list[2] - self.bbox_y_list[1])
            self.displacement_center_y.append(self.bbox_y_list[3] - self.bbox_y_list[2])
            self.displacement_center_y.append(self.bbox_y_list[4] - self.bbox_y_list[3])
            me_y = 1/4*(self.displacement_center_y[0] + self.displacement_center_y[1] + self.displacement_center_y[2] + \
                        self.displacement_center_y[3])
            if (self.displacement_center_y[3] > 0 and self.displacement_center_y[0] > 0 and self.displacement_center_y[
                1] > 0 and self.displacement_center_y[2] > 0 and me_x > 0) or (
                    self.displacement_center_y[3] < 0 and self.displacement_center_y[0] < 0 and
                    self.displacement_center_y[1] < 0 and self.displacement_center_y[2] < 0 and me_x < 0):
                self.displacement_mean_y = me_y
            else:
                self.displacement_mean_y = 0

        return self.bbox, maxscore, scale_x, fore_pixels5, fore_pixels9, p_95

    def detect_track(self, cropImg, pos, scale, penalty):
        score = []
        regression = []
        # pos = []
        # scale = []
        box_pred = []
        im_patch = self.transforms(cropImg)[None, :, :, :]
        search_region_output = self.model.search_region_feature_extract(im_patch.cuda())
        score_pred, center, delta, score_merge, best_pscore_id, bbox_pred = self.GMM_tracking(im_patch,
                                                                                              search_region_output)

        # score_pred, center, delta, score_merge, best_pscore_id, box

        maxscore = score_merge[best_pscore_id].mean() * penalty
        box = bbox_pred[best_pscore_id]
        box = np.mean(box, axis=0)
        if maxscore > self.thresh:
            # print('track maxscore ', maxscore)
            # print('penalty: ', penalty)
            score.append(maxscore)
            box_pred = box  # [sort_index]
            target = box_pred / scale  # 回到原图尺寸，中心和宽高

            # self.pos[0] = 上一帧绝对中心
            # target是以图像中心为原点的坐标
            res_x = target[0] + pos[0]
            res_y = target[1] + pos[1]
            lr = config.lr_box
            # config.min_scale = 0.1， config.min_scale = 10
            res_w = np.clip(self.bbox[2] * (1 - lr) + target[2] * lr, config.min_scale * self.origin_target_sz[0],
                            config.max_scale * self.origin_target_sz[0])
            res_h = np.clip(self.bbox[3] * (1 - lr) + target[3] * lr, config.min_scale * self.origin_target_sz[1],
                            config.max_scale * self.origin_target_sz[1])

            bbox = np.array([res_x, res_y, res_w, res_h])

            return bbox, maxscore
        else:
            return [], []

    def detect_track_occu(self, Img, scale):
        im_patch = self.transforms(Img)[None, :, :, :]
        search_region_output = self.model.search_region_feature_extract(im_patch.cuda())
        score_pred, center, delta, score_merge, best_pscore_id, bbox_pred = self.GMM_tracking(im_patch,
                                                                                             search_region_output)
        flag = 0
        pos = [self.bbox[0], self.bbox[1]]
        target = bbox_pred[best_pscore_id] / scale  # pos 回到原图尺寸，中心和宽高
        # self.pos[0] = 上一帧绝对中心
        # target是以图像中心为原点的坐标
        res_x = target[:, 0] + pos[0]
        res_y = target[:, 1] + pos[1] #

        # config.min_scale = 0.1， config.min_scale = 10
        lr = config.lr_box
        res_w = np.clip(self.bbox[2] * (1 - lr) + target[:, 2] * lr,
                        config.min_scale * self.origin_target_sz[0],
                        config.max_scale * self.origin_target_sz[0])
        res_h = np.clip(self.bbox[3] * (1 - lr) + target[:, 3] * lr,
                        config.min_scale * self.origin_target_sz[1],
                        config.max_scale * self.origin_target_sz[1])

        res_x = np.mean(res_x, axis=-1)
        res_y = np.mean(res_y, axis=-1)
        res_w = np.mean(res_w, axis=-1)
        res_h = np.mean(res_h, axis=-1)
        bbox = np.array([res_x, res_y, res_w, res_h])

        maxscore = score_merge[best_pscore_id].mean()
        if maxscore > self.thresh:
            # print('track maxscore:', maxscore)
            self.pos = np.array([res_x, res_y])
            self.target_sz = np.array([res_w, res_h])
            self.bbox = bbox
            flag = 1
        if flag == 0:
            self.pos = np.array([self.pos[0]+self.displacement_mean_x, self.pos[1]+self.displacement_mean_y])
            self.bbox = np.array([self.bbox[0] + self.displacement_mean_x, self.bbox[1] + self.displacement_mean_y, self.bbox[2], self.bbox[3]])
        # return self.bbox

    def redetect(self, frame): #, special=False
        # print('begin redetect-------')
        ###yolo目标检测###
        # 边缘消失
        h, w, _ = frame.shape
        cx, cy, cw, ch = self.bbox  # float type
        wc_z = cw + config.context_amount * (cw + ch)
        hc_z = ch + config.context_amount * (cw + ch)
        s_z = np.sqrt(wc_z * hc_z)  # the width of the crop box
        s_x = s_z * config.instance_size / config.exemplar_size  # sx 为sz 的两倍，为应该切的区域大小
        # print(self.bbox)
        if (cy-ch/2) < config.edge or h-config.edge < (cy+ch/2) or (cx-cw/2) < config.edge or w-config.edge < (cx+cw/2):
            # print('begin detect_track-------')
            step = int(s_x * 0.5)
            score = []
            bbox_pred = []
            num_wcut = w // step + 1
            num_hcut = h // step + 1
            if (cy-ch/2) < config.edge or self.miss_num > 20:
                for n in range(num_wcut):
                    if self.miss_num <= 20:
                        bbox = np.array([n*step, cy, cw, ch])
                    else:
                        bbox = np.array([n * step, 0, cw, ch])
                    d_c = (abs(cx - bbox[0])+ abs(cy - bbox[1]))/ (self.miss_num + 1)
                    penalty = np.exp(-d_c * config.dc_k)
                    cropImg, _, _, scale_x = get_instance_image(frame, bbox, config.exemplar_size,
                                                                        config.instance_size, config.context_amount,
                                                                        self.img_mean)
                    # cv2.imshow('1', cropImg)
                    # cv2.waitKeyEx(0)
                    pos = [bbox[0], bbox[1]]
                    box_pred, pscore = self.detect_track(cropImg, pos, scale_x, penalty)
                    if box_pred != []:
                        bbox_pred.append(box_pred)
                        score.append(pscore)
            if h - config.edge < (cy+ch/2) or self.miss_num > 20:
                for n in range(num_wcut):
                    if self.miss_num <= 20:
                        bbox = np.array([n*step, cy, cw, ch])
                    else:
                        bbox = np.array([n * step, h, cw, ch])
                    d_c = (abs(cx - bbox[0])+ abs(cy - bbox[1])) / (self.miss_num + 1)
                    penalty = np.exp(-d_c * config.dc_k)
                    cropImg, _, _, scale_x = get_instance_image(frame, bbox, config.exemplar_size,
                                                                        config.instance_size, config.context_amount,
                                                                        self.img_mean)
                    # cv2.imshow('1', cropImg)
                    # cv2.waitKeyEx(0)
                    pos = [bbox[0], bbox[1]]
                    box_pred, pscore = self.detect_track(cropImg, pos, scale_x, penalty)
                    if box_pred != []:
                        bbox_pred.append(box_pred)
                        score.append(pscore)

            if (cx-cw/2) < config.edge or self.miss_num > 20:
                for n in range(num_hcut):
                    if self.miss_num <= 20:
                        bbox = np.array([cx, n * step, cw, ch])
                    else:
                        bbox = np.array([0, n * step, cw, ch])
                    d_c = (abs(cx - bbox[0])+ abs(cy - bbox[1])) / (self.miss_num + 1)
                    penalty = np.exp(-d_c * config.dc_k)
                    cropImg, _, _, scale_x = get_instance_image(frame, bbox, config.exemplar_size,
                                                                config.instance_size, config.context_amount,
                                                                self.img_mean)
                    # cv2.imshow('1', cropImg)
                    # cv2.waitKeyEx(0)
                    pos = [bbox[0], bbox[1]]
                    box_pred, pscore = self.detect_track(cropImg, pos, scale_x, penalty)
                    if box_pred != []:
                        bbox_pred.append(box_pred)
                        score.append(pscore)
            if w - config.edge < (cx + cw/2) or self.miss_num > 20:
                for n in range(num_hcut):
                    if self.miss_num <= 20:
                        bbox = np.array([cx, n * step, cw, ch])
                    else:
                        bbox = np.array([w, n * step, cw, ch])
                    d_c = (abs(cx - bbox[0])+ abs(cy - bbox[1])) / (self.miss_num + 1)
                    penalty = np.exp(-d_c * config.dc_k)
                    cropImg, _, _, scale_x = get_instance_image(frame, bbox, config.exemplar_size,
                                                                config.instance_size, config.context_amount,
                                                                self.img_mean)
                    # cv2.imshow('1', cropImg)
                    # cv2.waitKeyEx(0)
                    pos = [bbox[0], bbox[1]]
                    box_pred, pscore = self.detect_track(cropImg, pos, scale_x, penalty)
                    if box_pred != []:
                        bbox_pred.append(box_pred)
                        score.append(pscore)

            if bbox_pred:
                id = np.argmax(score)
                self.bbox = bbox_pred[id]

        # 目标遮挡
        else:
            # print('begin detect_track_occu-------')
            instance_size = config.instance_size + self.miss_num * config.extend
            instance_img_np, _, _, scale_x = get_instance_image_detect(frame, self.bbox, config.exemplar_size,
                                                                       config.instance_size, instance_size,
                                                                       config.context_amount,
                                                                       self.img_mean)
            # scale_x = config.instance_size / instance_size * scale_x
            # instance_img_np = cv2.resize(instance_img_np, (config.instance_size, config.instance_size))
            self.detect_track_occu(instance_img_np, scale_x)

        return self.bbox




